"""

Package `StructuredArrays` implements arrays whose elements have values that
only depend on the indices. Example of such arrays are uniform arrays with
constant values, structured arrays with boolean values indicating whether an
entry is significant.

"""
module StructuredArrays

export
    AbstractStructuredArray,
    AbstractStructuredMatrix,
    AbstractStructuredVector,
    AbstractUniformArray,
    AbstractUniformMatrix,
    AbstractUniformVector,
    FastUniformArray,
    FastUniformMatrix,
    FastUniformVector,
    MutableUniformArray,
    MutableUniformMatrix,
    MutableUniformVector,
    StructuredArray,
    StructuredMatrix,
    StructuredVector,
    UniformArray,
    UniformMatrix,
    UniformVector

using TypeUtils

import Base: @propagate_inbounds, front, tail, to_indices

const SubArrayRange = Union{Integer,OrdinalRange{<:Integer,<:Integer},Colon}
const ConcreteIndexStyle = Union{IndexLinear,IndexCartesian}

# Type of array axes and union of array size or axes.
const Axes{N} = NTuple{N,AbstractUnitRange{Int}}
const DimsOrAxes{N} = Union{Dims{N},Axes{N}}

abstract type AbstractStructuredArray{T,N,S<:ConcreteIndexStyle,I<:DimsOrAxes{N}} <: AbstractArray{T,N} end

abstract type AbstractUniformArray{T,N,I<:DimsOrAxes{N}} <: AbstractStructuredArray{T,N,IndexLinear,I} end

"""
    UniformArray(val, args...) -> A
    UniformArray{T}(val, args...) -> A
    UniformArray{T,N}(val, args...) -> A

build an immutable array `A` whose elements are all equal to `val`. The storage
requirement is `O(1)` instead of `O(length(A))` for a usual array `A`.
Subsequent arguments `args...` define the size or the axes of the array.
Optional parameters `T` and `N` are to specify the element type and the number
of dimensions of the array.

Uniform arrays implement conventional linear indexing: `A[i]` yields `val` for
all linear indices `i` in the range `1:length(A)`.

A statement like `A[i] = val` is not implemented as uniform arrays are
considered as immutable. Call `MutableUniformArray(val,dims)` to create a
uniform array whose element value can be changed.

"""
struct UniformArray{T,N,I<:DimsOrAxes{N}} <: AbstractUniformArray{T,N,I}
    inds::I
    val::T
    UniformArray{T}(val, inds::I) where {T,N,I<:DimsOrAxes{N}} =
        new{T,N,I}(checked_size(inds), val)
end

"""
    FastUniformArray(val, args...) -> A
    FastUniformArray{T}(val, args...) -> A
    FastUniformArray{T,N}(val, args...) -> A

build an immutable uniform array `A` whose elements are all equal to `val`. The
difference with an instance of [`UniformArray`](@ref) is that `val` is part of
the type signature so that `val` can be known at compile time. A typical use is
to create all true/false masks.

"""
struct FastUniformArray{T,N,V,I<:DimsOrAxes{N}} <: AbstractUniformArray{T,N,I}
    inds::I
    FastUniformArray{T}(val, inds::I) where {T,N,I<:DimsOrAxes{N}} =
        new{T,N,as(T,val),I}(checked_size(inds))
end

"""
    MutableUniformArray(val, args...) -> A
    MutableUniformArray{T}(val, args...) -> A
    MutableUniformArray{T,N}(val, args...) -> A

build a mutable array `A` whose elements are initially all equal to `val`. The
difference with an instance of [`UniformArray`](@ref) is that the uniform value
can be changed.

A statement like `A[i] = val` is allowed but changes the value of all the
elements of `A`. Call `UniformArray(val,dims)` to create an immutable uniform
array whose element value cannot be changed.

"""
mutable struct MutableUniformArray{T,N,I<:DimsOrAxes{N}} <: AbstractUniformArray{T,N,I}
    inds::I
    val::T
    MutableUniformArray{T}(val, inds::I) where {T,N,I<:DimsOrAxes{N}} =
        new{T,N,I}(checked_size(inds), val)
end

"""
    StructuredArray([S = IndexCartesian,] func, args...) -> A
    StructuredArray{T}([S = IndexCartesian,] func, args...) -> A
    StructuredArray{T,N}([S = IndexCartesian,] func, args...) -> A
    StructuredArray{T,N,S}(func, args...) -> A

build an array `A` whose values are a given function, here `func`, of its
indices: `A[i]` is computed as `func(i)`. The storage requirement is `O(1)`
instead of `O(lenght(A))` for a usual array. Subsequent arguments `args...`
define the size or the axes of the array. Optional parameters `T`, `N`, and `S`
are to specify the element type, the number of dimensions, and the type of the
indexing style of the array. If specified as an argument not a type parameter,
`S` may also be an instance of `IndexStyle`.

If indexing style is `IndexCartesian` (the default), the function `func` will
be called with `N` integer arguments, `N` being the number of dimensions. If
`S` is `IndexLinear`, the function `func` will be called with a single integer
argument.

For instance, the structure of a lower triangular matrix of size `m×n` would be
given by:

    StructuredArray((i,j) -> (i ≥ j), m, n)

but with a constant small storage requirement whatever the size of the matrix.

Although the callable object `func` may not be a *pure function*, its return
type shall be stable and structured arrays are considered as immutable in the
sense that a statement like `A[i] = val` is not implemented. The type of the
elements of structured array is guessed by applying `func` to the unit index.
The element type, say `T`, may also be explicitely specified:

    StructuredArray{T}([S = IndexCartesian,] func, dims)

"""
struct StructuredArray{T,N,S,F,I<:DimsOrAxes{N}} <: AbstractStructuredArray{T,N,S,I}
    inds::I
    func::F
    StructuredArray{T,N,S}(func::F, inds::I) where {T,N,S<:ConcreteIndexStyle,F,I<:DimsOrAxes{N}} =
        new{T,N,S,F,I}(checked_size(inds), func)
end

# Aliases.
for (A, N) in ((:Vector, 1), (:Matrix, 2))
    @eval begin
        const $(Symbol("AbstractStructured",A)){T,S,I} = AbstractStructuredArray{T,$N,S,I}
        const $(Symbol("AbstractUniform",A)){T,I} = AbstractUniformArray{T,$N,I}
        const $(Symbol("MutableUniform",A)){T,I} = MutableUniformArray{T,$N,I}
        const $(Symbol("Uniform",A)){T,I} = UniformArray{T,$N,I}
        const $(Symbol("FastUniform",A)){T,V,I} = FastUniformArray{T,$N,V,I}
        const $(Symbol("Structured",A)){T,S,F,I} = StructuredArray{T,$N,S,F,I}
    end
end

# Specialize base abstract array methods for structured arrays.
for cls in (:StructuredArray, :FastUniformArray, :UniformArray, :MutableUniformArray)
    @eval begin
        Base.length(A::$cls) = prod(size(A))
        Base.size(A::$cls) = to_size(getfield(A, :inds))
        Base.size(A::$cls, i::Integer) =
            i > ndims(A) ? 1 :
            i > zero(i) ? to_dim(getfield(A, :inds)[i]) : throw(BoundsError(size(A), i))
        Base.axes(A::$cls) = to_axes(getfield(A, :inds))
        Base.axes(A::$cls, i::Integer) =
            i > ndims(A) ? Base.OneTo(1) :
            i > zero(i) ? to_axis(getfield(A, :inds)[i]) : throw(BoundsError(axes(A), i))
     end
end
Base.has_offset_axes(A::AbstractUniformArray{T,0,Tuple{}}) where {T} = false
Base.has_offset_axes(A::AbstractUniformArray{T,N,Dims{N}}) where {T,N} = false
Base.has_offset_axes(A::StructuredArray{T,0,S,F,Tuple{}}) where {T,S,F} = false
Base.has_offset_axes(A::StructuredArray{T,N,S,F,Dims{N}}) where {T,N,S,F} = false

# Constructors that convert trailing argument(s) to array dimensions or axes.
for cls in (:StructuredArray, :FastUniformArray, :UniformArray, :MutableUniformArray)
    # 0-dimesional case.
    @eval begin
        $cls(     arg1)             = $cls(     arg1, ())
        $cls{T}(  arg1) where {T}   = $cls{T}(  arg1, ())
        $cls{T,N}(arg1) where {T,N} = $cls{T,N}(arg1, ())
    end
    if cls === :StructuredArray
        @eval begin
            $cls(::Union{S,Type{S}}, func) where {S<:ConcreteIndexStyle} =
                $cls(S, func, ())
            $cls{T}(::Union{S,Type{S}}, func) where {T,S<:ConcreteIndexStyle} =
                $cls{T}(S, func, ())
            $cls{T,0}(::Union{S,Type{S}}, func) where {T,S<:ConcreteIndexStyle} =
                $cls{T,0,S}(func, ())
            $cls{T,0,S}(::Union{S,Type{S}}, func) where {T,S<:ConcreteIndexStyle} =
                $cls{T,0,S}(func, ())
            $cls{T,0,S}(func) where {T,S<:ConcreteIndexStyle} =
                $cls{T,0,S}(func, ())
        end
    end
    # N-dimensional cases with N ≥ 1.
    for (type, to) in ((Integer,                      :to_size),
                       (AbstractUnitRange{<:Integer}, :to_axes))
        @eval begin
            $cls(     arg1, args::$type...)             = $cls(     arg1, args)
            $cls{T}(  arg1, args::$type...) where {T}   = $cls{T}(  arg1, args)
            $cls{T,N}(arg1, args::$type...) where {T,N} = $cls{T,N}(arg1, args)

            $cls(     arg1, args::NTuple{N,$type}) where {  N} = $cls(     arg1, $to(args))
            $cls{T}(  arg1, args::NTuple{N,$type}) where {T,N} = $cls{T}(  arg1, $to(args))
            $cls{T,N}(arg1, args::NTuple{N,$type}) where {T,N} = $cls{T,N}(arg1, $to(args))
        end
        if cls === :StructuredArray
            @eval begin
                $cls(::Union{S,Type{S}}, func, args::$type...) where {S<:ConcreteIndexStyle} =
                    $cls(S, func, args)
                $cls{T}(::Union{S,Type{S}}, func, args::$type...) where {T,S<:ConcreteIndexStyle} =
                    $cls{T}(S, func, args)
                $cls{T,N}(::Union{S,Type{S}}, func, args::$type...) where {T,N,S<:ConcreteIndexStyle} =
                    $cls{T,N,S}(func, args)
                $cls{T,N,S}(::Union{S,Type{S}}, func, args::$type...) where {T,N,S<:ConcreteIndexStyle} =
                    $cls{T,N,S}(func, args)
                $cls{T,N,S}(func, args::$type...) where {T,N,S<:ConcreteIndexStyle} =
                    $cls{T,N,S}(func, args)

                $cls(::Union{S,Type{S}}, func, args::NTuple{N,$type}) where {N,S<:ConcreteIndexStyle} =
                    $cls(S, func, $to(args))
                $cls{T}(::Union{S,Type{S}}, func, args::NTuple{N,$type}) where {T,N,S<:ConcreteIndexStyle} =
                    $cls{T}(S, func, $to(args))
                $cls{T,N}(::Union{S,Type{S}}, func, args::NTuple{N,$type}) where {T,N,S<:ConcreteIndexStyle} =
                    $cls{T,N,S}(func, $to(args))
                $cls{T,N,S}(::Union{S,Type{S}}, func, args::NTuple{N,$type}) where {T,N,S<:ConcreteIndexStyle} =
                    $cls{T,N,S}(func, $to(args))
                $cls{T,N,S}(func, args::NTuple{N,$type}) where {T,N,S<:ConcreteIndexStyle} =
                    $cls{T,N,S}(func, $to(args))
            end
        end
    end
end

# Default index style is Cartesian for StructuredArray.
let cls = :StructuredArray
    @eval begin
        $cls(func, args::DimsOrAxes{N}) where {N} = $cls(IndexCartesian, func, args)
        $cls{T}(func, args::DimsOrAxes{N}) where {T,N} = $cls{T,N}(IndexCartesian, func, args)
        $cls{T,N}(func, args::DimsOrAxes{N}) where {T,N} = $cls{T,N}(IndexCartesian, func, args)
    end
end

# Constructors that manage to call the inner constructors.
for cls in (:FastUniformArray, :UniformArray, :MutableUniformArray)
    @eval begin
        $cls(val::T, args::DimsOrAxes{N}) where {T,N} = $cls{T}(val, args)
        $cls{T,N}(val, args::DimsOrAxes{N}) where {T,N} = $cls{T}(val, args)
    end
end
let cls = :StructuredArray
    @eval begin
        function $cls(::Union{S,Type{S}}, func, args::DimsOrAxes{N}) where {N,S<:ConcreteIndexStyle}
            T = guess_eltype(func, S, Val(N))
            return $cls{T,N,S}(func, args)
        end
        $cls{T}(::Union{S,Type{S}}, func, args::DimsOrAxes{N}) where {T,N,S<:ConcreteIndexStyle} =
            $cls{T,N,S}(func, args)
        $cls{T,N}(::Union{S,Type{S}}, func, args::DimsOrAxes{N}) where {T,N,S<:ConcreteIndexStyle} =
            $cls{T,N,S}(func, args)
    end
end

guess_eltype(func, ::Type{IndexLinear}, ::Val{N}) where {N} =
    Base.promote_op(func, Int)

@generated guess_eltype(func, ::Type{IndexCartesian}, ::Val{N}) where {N} =
    :(Base.promote_op(func, $(ntuple(x -> Int, Val(N))...)))

# Specialize some methods for (fast) uniform arrays of booleans.
Base.all(A::AbstractUniformArray{Bool}) = value(A)
Base.count(A::AbstractUniformArray{Bool}) = ifelse(value(A), length(A), 0)
Base.count(A::FastUniformArray{Bool,N,true}) where {N} = length(A)
Base.count(A::FastUniformArray{Bool,N,false}) where {N} = 0

# See comments near `getindex` in `abstractarray.jl` for explanations about how
# `getindex` and `setindex!` methods are expected to be specialized depending
# on the indexing style: `i:Int` for linear indexing style or
# `i::Vararg{Int,N}` for `N` dimensional Cartesian indexing style.

Base.IndexStyle(::Type{<:AbstractStructuredArray{T,N,S}}) where {T,N,S} = S()

@inline function Base.getindex(A::StructuredArray{T,N,IndexLinear},
                               i::Int) where {T,N}
    @boundscheck checkbounds(A, i)
    return @inbounds as(T, A.func(i))
end

@inline function Base.getindex(A::StructuredArray{T,N,IndexCartesian},
                               I::Vararg{Int,N}) where {T,N}
    @boundscheck checkbounds(A, I...)
    return @inbounds as(T, A.func(I...))
end

"""
    StructuredArrays.value(A::AbstractUniformArray)

yields the value of the elements of the uniform array `A`.

"""
value(A::FastUniformArray{T,N,V}) where {T,N,V} = V
value(A::UniformArray) = getfield(A, :val)
value(A::MutableUniformArray) = getfield(A, :val)

setvalue!(A::MutableUniformArray{T}, x) where {T} = setfield!(A, :val, as(T, x))

@inline function Base.getindex(A::AbstractUniformArray, i::Int)
    @boundscheck checkbounds(A, i)
    return value(A)
end

# Fast view for immutable uniform arrays, may return a 0-dimensional array.
@inline @propagate_inbounds function Base.view(A::Union{UniformArray,FastUniformArray},
                                               I::Vararg{Any})
    return subarray(A, I...)
end

@inline function Base.setindex!(A::MutableUniformArray, x, i::Int)
    i:i == eachindex(IndexLinear(), A) || not_all_elements()
    setvalue!(A, x)
    return A
end

@inline function Base.setindex!(A::MutableUniformArray, x, rng::AbstractUnitRange{<:Integer})
    rng == eachindex(IndexLinear(), A) || not_all_elements()
    setvalue!(A, x)
    return A
end

@inline function Base.setindex!(A::MutableUniformArray, x, ::Colon)
    setvalue!(A, x)
    return A
end

@noinline not_all_elements() =
    error("all elements must be set at the same time")

@inline function subarray(A::AbstractUniformArray, I::Vararg{Any})
    # NOTE: Since we are calling the `subarray` method to produce an object
    # that is independent from `A`, it is not needed to unalias the result of
    # `to_indices` from `A` as is done in Julia base code (in `subarray.jl`) to
    # build views.
    J = to_indices(A, I)
    @boundscheck checkbounds(A, J...)
    # NOTE: `to_indices` converts colons (`:`) into slices (`Base.Slice`) which
    # are abstract unit ranges and thus abstract vectors of indices. So the
    # only thing to do is to concatenate the sizes of the indices returned by
    # `to_indices`.
    dims = concat(size, J)
    return parameterless(typeof(A))(value(A), concat(size, J))
end

@inline concat(f::Function, x::Tuple) = concat((), f, x) # start recursion
@inline concat(r::Tuple, f::Function, x::Tuple{}) = r    # finish recursion
@inline concat(r::Tuple, f::Function, x::Tuple) =        # pursue recursion
    concat((r..., f(first(x))...), f, tail(x))

"""
    StructuredArrays.checked_size(inds) -> inds

throws an `ArgumentError` exception if `inds` is not a valid array size
or array axes amd returns `inds` otherwise.

"""
function checked_size(dims::Dims{N}) where {N}
    flag = true
    @inbounds for i in 1:N
        flag &= dims[i] ≥ 0
    end
    flag || throw(ArgumentError("invalid array dimension(s)"))
    return dims
end
checked_size(inds::Axes) = inds
checked_size(::Tuple{}) = ()

to_dim(dim::Int) = dim
to_dim(dim::Integer) = Int(dim)
to_dim(rng::AbstractUnitRange{<:Integer}) = length(rng)::Int

to_size(::Tuple{}) = ()
to_size(dims::Dims) = dims
to_size(dims::Tuple{Vararg{Integer}}) = map(to_dim, dims)
to_size(dim::Integer) = (to_dim(dim),)
to_size(inds::Tuple{Vararg{AbstractUnitRange{<:Integer}}}) = map(length, inds)

to_axis(rng::AbstractUnitRange{Int}) = rng
to_axis(rng::AbstractUnitRange{<:Integer}) = convert_eltype(Int, rng)
to_axis(dim::Integer) = Base.OneTo{Int}(dim)

to_axes(::Tuple{}) = ()
to_axes(inds::Tuple{Vararg{AbstractUnitRange{Int}}}) = inds
to_axes(inds::Tuple{Vararg{AbstractUnitRange{<:Integer}}}) = map(to_axis, inds)
to_axes(dims::Tuple{Vararg{Integer}}) = map(to_axis, dims)

end # module
